extends Node

signal slot_assigned

const HealingVfx = [preload("res://data/hit_vfx/status_healing_1.tres")]

export  var team:int = 0

onready var battle = get_parent()

var status:StatusNode
var slot:BattleSlot
var status_bubble:StatusBubble
var current_decoy:Decoy
var character_sfx_override:CharacterSfx
var hide_from_party:bool

var had_turn:bool = false
var turn_num:int = 0
var damage_this_round:int = 0



var last_used_move = null



var next_tape = null


var next_item = null



var will_fuse:bool = false




var just_desynced:bool = false


var last_damage_source:WeakRef = weakref(null)

func _init():
	status = StatusNode.new()
	add_child(status)

func is_player_controlled():
	return get_controller() is PlayerFighterController

func is_remote_player_controlled():
	return get_controller() is RemotePlayerFighterController

func get_net_player_id():
	var request = battle.get_net_request()
	if not request:
		return null
	if is_player_controlled():
		return request.local_id
	elif is_remote_player_controlled():
		return request.remote_id
	return null

func get_controller():
	for child in get_children():
		if child is FighterController:
			return child
	assert (false)
	return null

func get_characters()->Array:
	var result = []
	for child in get_children():
		if child is CharacterNode:
			result.push_back(child)
	return result

func has_character(c:Resource)->bool:
	for child in get_children():
		if child is CharacterNode:
			if child.character == c:
				return true
	return false

func get_character_kind()->int:
	return get_controller().get_character_kind()

func get_character_sfx()->CharacterSfx:
	if is_fusion():
		return null
	if character_sfx_override:
		return character_sfx_override
	for child in get_children():
		if child is CharacterNode and child.character and child.character.sfx:
			return child.character.sfx
	return null

func get_species()->Array:
	var result = []
	for child in get_children():
		if child is CharacterNode and child.active_tape != null:
			result.push_back(child.active_tape.form)
	return result

func is_transformed()->bool:
	for child in get_children():
		if child is CharacterNode and child.active_form != null:
			return true
	return false

func is_fusion()->bool:
	return get_characters().size() >= 2

func is_recordable()->bool:
	if is_fusion() or status.dead or status.has_tag("recorded"):
		return false
	for c in get_characters():
		if c.character.force_unrecordable:
			return false
		if Character.is_human(c.character.character_kind):
			return false
		if c.character.character_kind == Character.CharacterKind.ARCHANGEL:
			return false
	for f in battle.get_teams()[team]:
		
		if Character.is_human(f.get_character_kind()):
			return false
	return true

func can_record()->bool:
	return not is_fusion() and not will_fuse and not status.has_tag("sleep")

func can_fuse_with(other_fighter)->bool:
	return get_controller().can_fuse_with(other_fighter)

func get_fuser()->Node:
	
	
	var team = battle.get_teams(false, false)[self.team]
	for f in team:
		if f != self and can_fuse_with(f) and f.can_fuse_with(self):
			return f
	return null

func is_fusion_unlocked()->bool:
	return get_controller().is_fusion_unlocked()

func is_fusion_power_unlocked()->bool:
	return get_controller().is_fusion_power_unlocked()

func get_fusion_power_seed()->int:
	var fusion_seed = 0
	for child in get_characters():
		if child.active_tape != null:
			fusion_seed = (fusion_seed << 16 | fusion_seed >> 16)
			fusion_seed ^= hash(child.active_tape.form.name)
	return fusion_seed
	
func get_fusion_seed()->int:
	
	return 0

func is_bootleg()->bool:
	for c in get_characters():
		if c.current_tape and c.current_tape.type_override.size() > 0:
			return true
	return false

func get_general_form()->BaseForm:
	return get_controller().get_general_form()

func get_ap_system()->int:
	return get_controller().get_ap_system()

func get_character_name()->String:
	var characters = get_characters()
	if characters.size() == 1:
		if characters[0].form_name != "":
			match DLC.mods_by_id.use_tape_name_by_ashley_crow.setting_player_name:
				0:
					return tr(characters[0].form_name)
				1:
					return tr(characters[0].form_name) + " (" + tr(characters[0].character_name) + ")"
				2:
					return tr(characters[0].character_name)
				3:
					return tr(characters[0].character_name) + " (" + tr(characters[0].form_name) + ")"
		return tr(characters[0].character_name)
	var name_parts = []
	var any_char_names = false
	for character in characters:
		if character.character_name != "":
			match DLC.mods_by_id.use_tape_name_by_ashley_crow.setting_fusion_name:
				0:
					name_parts.push_back(tr(character.form_name))
				1:
					name_parts.push_back(tr(character.form_name) + " (" + tr(character.character_name) + ")")
				2:
					name_parts.push_back(tr(character.character_name))
				3:
					name_parts.push_back(tr(character.character_name) + " (" + tr(character.form_name) + ")")
			any_char_names = true
		else :
			name_parts.push_back(tr(character.form_name))
	if not any_char_names:
		return ""
	return Strings.join("/", name_parts)

func _get_char_pronouns(char_node)->int:
	if _refer_to_by_char_name():
		return char_node.get_character_pronouns()
	return char_node.get_form_pronouns()

func get_pronouns()->int:
	var characters = get_characters()
	if characters.size() == 0:
		return Loc.Pronouns.THEY
	if characters.size() == 1:
		return _get_char_pronouns(characters[0])
	for i in range(1, characters.size()):
		if _get_char_pronouns(characters[i]) != _get_char_pronouns(characters[0]):
			return Loc.Pronouns.THEY
	return _get_char_pronouns(characters[0])

func get_form_name()->String:
	if is_fusion():
		var form = get_general_form()
		if form != null:
			return tr(form.name)
	var name_parts = []
	for character in get_characters():
		if character.form_name != "":
			name_parts.push_back(tr(character.form_name))
		else :
			name_parts.push_back(tr(character.character_name))
	return Strings.join("/", name_parts)

func get_team_name_override()->String:
	for character in get_characters():
		if character.character.team_name_override != "":
			return character.character.team_name_override
	return ""

func get_flee_chance()->Rational:
	if not battle.can_flee:
		return Rational.new(0, 1)
	var total_speed:int = 0
	var total_evasion:int = 0
	var enemy_speed:int = 0
	var enemy_accuracy:int = 0
	
	for f in battle.get_fighters(false):
		if f.team == team:
			total_speed += f.status.speed
			total_evasion += f.status.evasion
		elif not f.status.has_tag("ghostly"):
			enemy_speed += f.status.speed
			enemy_accuracy += f.status.accuracy
	
	return BattleFormulas.get_flee_chance(total_speed, total_evasion, enemy_speed, enemy_accuracy)

func _refer_to_by_char_name()->bool:
	return Character.is_human(get_character_kind()) and (team == 0 or get_team_name_override() != "")

func get_general_name()->String:
	if _refer_to_by_char_name():
		return get_character_name()
	else :
		return get_form_name()

func get_modified_name()->String:
	return status.modify_name(get_general_name(), get_pronouns())

func get_disambiguator()->String:
	var disambig = battle.name_disambig.get_disambig(self, get_general_name())
	if battle.name_disambig.count_disambigs(self, get_general_name()) <= 1:
		return ""
	return " #" + str(disambig)

func get_name_with_disambiguator()->String:
	return get_general_name() + get_disambiguator()

func get_name_with_team()->String:
	var team_name = get_team_name_override()
	if team_name == "":
		team_name = "BATTLE_ADJECTIVE_TEAM" + str(team)
	return status.modify_name(Loc.trgf(team_name, get_pronouns(), {
		"name":get_general_name(), 
		"disambiguator":get_disambiguator()
	}), get_pronouns())

func get_moves()->Array:
	return status.get_moves()

func get_loot_table()->Resource:
	for child in get_characters():
		if child.active_form and child.active_form.loot_table != "":
			return load(child.active_form.loot_table)
	for type in status.get_types():
		if type.loot_table != "":
			return load(type.loot_table)
	return null

func add_status_bubble(slot, snapshot):
	if status_bubble or not slot.status_bubble_enabled:
		return 
	status_bubble = battle.setup_status_bubble(slot)
	status_bubble.set_snapshot(snapshot, false)
	status_bubble.fade_in()

func set_preferred_slot():
	slot = null
	for character in get_characters():
		if character.slot != null:
			slot = character.slot
			emit_signal("slot_assigned")
			return 

func animate_leave_slot(old_slot):
	assert (old_slot != null)
	
	for effect in status.get_effects():
		effect.effect.leave_slot(effect, old_slot)
	
	old_slot.clear()

func _animate_enter_slot(new_slot, visuals):
	assert (new_slot != null)
	new_slot.reset_position(true)
	new_slot.set_form(visuals.form)
	new_slot.set_palette(visuals.palette, false)
	new_slot.set_sparkle(visuals.sparkle)
	if not visuals.transformed:
		new_slot.set_idle_override("inactive")
		new_slot.play_animation("inactive")
	
	get_controller()._animate_after_enter_slot(new_slot)
	
	for effect in status.get_effects():
		effect.effect.enter_slot(effect, new_slot)

func _get_visuals()->Dictionary:
	return {
		form = get_general_form(), 
		palette = get_palette(), 
		sparkle = get_sparkle(), 
		transformed = is_transformed()
	}

func generate_enter_slot_animation()->Bind:
	assert (slot != null)
	return Bind.new(self, "_animate_enter_slot", [slot, _get_visuals()])

func _animate_transform(slot, visuals, snap, idle_override:String = "", voice = null):
	assert (slot != null)
	if voice != null:
		slot.play_voice(voice)
	if slot.idle_override == "inactive":
		yield (slot.wait_for_idle_sync(), "completed")
		slot.set_idle_override("")
		yield (slot.play_sprite_animation("transform"), "completed")
	yield (slot.transform_to(visuals.form, visuals.palette, visuals.sparkle, idle_override), "completed")
	if status_bubble:
		status_bubble.set_snapshot(snap)

func generate_transform_animation(first_time:bool = false)->Bind:
	assert (slot != null)
	var voice = null
	if first_time:
		var sfx = get_character_sfx()
		if sfx:
			voice = sfx.transform
	return Bind.new(self, "_animate_transform", [slot, _get_visuals(), status.get_snapshot(), "", voice])

func get_palette()->Array:
	return status.get_palette()

func get_sparkle()->bool:
	for type in status.get_types():
		if type.sparkle:
			return true
	return false

func animate_hit(slot, damage:Damage, snapshot = null):
	var co_list = [
		animate_hit_vfx(slot, damage), 
		Co.delayed(damage.hit_delay, Bind.new(self, "_animate_hit_hit", [slot, damage, snapshot]))]
	if not damage.is_splash:
		yield (Co.join(co_list), "completed")

func _animate_hit_hit(slot, damage:Damage, snapshot = null):
	if status_bubble:
		status_bubble.shake()
	get_controller().damage_hit_vibrate(damage)
	battle.camera_set_state("damage", [slot])
	battle.camera_set_state("normal", [slot])
	
	var toast = battle.create_toast()
	toast.setup_damage(damage)
	
	var co_list = [
		battle.play_toast(toast, slot), 
		status_bubble.set_snapshot(snapshot) if status_bubble != null and snapshot != null else null, 
		slot.play_animation(get_controller().get_hurt_animation_name())]
	if not damage.is_splash:
		yield (Co.join(co_list), "completed")

func animate_heal(slot, hp_healed:int, snapshot, toast_message:String = ""):
	battle.camera_set_state("heal", [slot])
	var toast = battle.create_toast()
	toast.setup_heal(hp_healed)
	toast.message = toast_message
	yield (Co.join([
		battle.play_toast(toast, slot), 
		animate_vfx_sequence(HealingVfx), 
		status_bubble.set_snapshot(snapshot) if status_bubble else null
	]), "completed")

func animate_miss(slot):
	battle.camera_set_state("missed", [slot])
	var toast = battle.create_toast()
	toast.setup_text("BATTLE_TOAST_MISSED")
	return battle.play_toast(toast, slot)

func animate_decoy_hit(slot, damage:Damage):
	get_controller().damage_hit_vibrate(damage)
	battle.camera_set_state("damage", [slot])
	battle.camera_set_state("normal", [slot])
	yield (Co.join([
		animate_hit_vfx(slot, damage, "decoy"), 
		Co.delayed(damage.hit_delay, Bind.new(slot, "play_decoy_animation", ["hurt"]))]), "completed")

func animate_hit_vfx(slot, damage:Damage, position_override = null):
	var target_slots = []
	if damage.source != null:
		target_slots.push_back(damage.source_slot)
	else :
		target_slots = battle.get_slots()
	
	var hit_vfx = damage.hit_vfx
	for vfx in hit_vfx:
		var scene = vfx.instance(battle.background, slot, target_slots, damage.types)
		scene.position_override = position_override
		battle.background.add_child(scene)
		yield (scene, "vfx_finished")
	if hit_vfx.size() == 0:
		return Co.pass()

func animate_vfx_sequence(vfx_seq:Array, elemental_types:Array = []):
	for vfx in vfx_seq:
		var scene = vfx.instance(battle.background, self.slot, battle.get_slots(), elemental_types)
		battle.background.add_child(scene)
		yield (scene, "vfx_finished")
	if vfx_seq.size() == 0:
		return Co.pass()

func animate_defeat_monster():
	battle.camera_set_state("defeat", [slot])
	yield (Co.join([
		
		status_bubble.animate_death(true) if status_bubble else Co.pass(), 
		slot.play_animation("flee")
	]), "completed")
	animate_leave_slot(slot)

func animate_defeat_ghost():
	battle.camera_set_state("defeat", [slot])
	yield (Co.join([
		
		status_bubble.animate_death(true) if status_bubble else Co.pass(), 
		slot.play_animation("ghostly_fade")
	]), "completed")
	animate_leave_slot(slot)
	
func animate_defeat_summon():
	battle.camera_set_state("defeat", [slot])
	yield (Co.join([
		
		status_bubble.animate_death(true) if status_bubble else Co.pass(), 
		slot.play_animation("flee")
	]), "completed")
	animate_leave_slot(slot)

func animate_defeat_human_tape(transform_anim:Bind, snapshot):
	battle.camera_set_state("defeat", [slot])
	yield (Co.join([
		
		status_bubble.set_snapshot(snapshot, true) if status_bubble else Co.pass(), 
		transform_anim.call_func()
	]), "completed")

func animate_defeat_human(transform_anim:Bind, _reason_no_tapes:bool):
	battle.camera_set_state("defeat", [slot])
	if not is_transformed():
		slot.set_idle_override("defeated")
		slot.play_slot_animation("knockdown")
	yield (Co.join([
		
		status_bubble.animate_death(true) if status_bubble else Co.pass(), 
		transform_anim.call_func1("defeated") if is_transformed() else slot.play_sprite_animation("defeated")
	]), "completed")

func animate_defeat_knockdown():
	battle.camera_set_state("defeat", [slot])
	slot.set_idle_override("defeated")
	slot.play_slot_animation("knockdown")
	yield (Co.join([
		
		status_bubble.animate_death(true) if status_bubble else Co.pass(), 
		slot.play_sprite_animation("defeated")
	]), "completed")

func set_decoy(decoy:Decoy):
	if decoy == current_decoy:
		return 
	battle.queue_animation(Bind.new(self, "_animate_decoy_change", [slot, current_decoy, decoy]))
	current_decoy = decoy

func _animate_decoy_change(slot, old_decoy:Decoy, new_decoy:Decoy):
	if old_decoy != null:
		slot.play_sound(old_decoy.despawn_sound)
		yield (slot.play_decoy_animation("spawn", true), "completed")
	slot.set_decoy(new_decoy.sprite if new_decoy != null else null)
	if new_decoy != null:
		slot.play_sound(new_decoy.spawn_sound)
		yield (slot.play_decoy_animation("spawn"), "completed")
